import logging

from sanic import Blueprint, response
from sanic.request import Request

from rasax.community import utils
from rasax.community.api.decorators import platform_scoped, inject_rasa_x_user
from rasax.community.services.nlg_service import NlgService
from rasax.community.services.projects_service import ProjectsService
from rasax.community.utils import error

logger = logging.getLogger(__name__)


def _nlg_service(request: Request) -> NlgService:
    return NlgService(request["db_session"])


def _project_service(request: Request) -> ProjectsService:
    return ProjectsService(request["db_session"])


def blueprint() -> Blueprint:
    nlg_endpoints = Blueprint('nlg_endpoints')

    @nlg_endpoints.route("/templates", methods=['GET', 'HEAD'])
    @platform_scoped('responseTemplates.list', allow_api_token=True)
    async def get_templates(request):
        text_query = utils.default_arg(request, 'q', None)
        template_query = utils.default_arg(request, 'template', None)
        fields = utils.fields_arg(
            request, {'text', 'template', 'id'}
        )

        limit = int(utils.default_arg(request, 'limit', 0)) or None
        offset = int(utils.default_arg(request, 'offset', 0)) or None

        templates, total_number = _nlg_service(request).fetch_templates(
            text_query, template_query, fields, limit, offset)

        return response.json(templates,
                             headers={"X-Total-Count": total_number})

    @nlg_endpoints.route("/templates", methods=['POST'])
    @platform_scoped('responseTemplates.create', allow_api_token=True)
    @inject_rasa_x_user(allow_api_token=True)
    async def add_template(request, user):
        rjs = request.json
        if not utils.check_schema("nlg/response", rjs):
            return error(400, "WrongSchema", "Wrong nlg response schema")

        domain_id = _project_service(request).get_domain_id()
        saved_response = _nlg_service(request).save_template(rjs, user["username"],
                                                             domain_id=domain_id)
        _project_service(request).dump_domain_in_local_mode()
        return response.json(saved_response)

    @nlg_endpoints.route("/templates", methods=['PUT'])
    @platform_scoped('bulkResponseTemplates.update', allow_api_token=True)
    @inject_rasa_x_user(allow_api_token=True)
    async def update_templates(request, user=None):
        """Delete old bot responses and replace them with the responses in the
        payload."""

        rjs = request.json
        if not utils.check_schema("nlg/response_bulk", rjs):
            return error(400, "WrongSchema",
                         "Schema for nlg response_bulk payload not recognised.")
        domain_id = _project_service(request).get_domain_id()
        inserted_count = _nlg_service(request).replace_templates(rjs, user["username"],
                                                                 domain_id=domain_id)
        _project_service(request).dump_domain_in_local_mode()
        return response.text(("Successfully uploaded {} responses."
                              ).format(inserted_count))

    @nlg_endpoints.route("/templates/<response_id>", methods=['PUT'])
    @platform_scoped('responseTemplates.update', allow_api_token=True)
    @inject_rasa_x_user(allow_api_token=True)
    async def modify_template(request, response_id, user):
        rjs = request.json
        if not utils.check_schema("nlg/response", rjs):
            return error(400, "WrongSchema", "Wrong nlg response schema")
        updated_response = _nlg_service(request).update_template(response_id,
                                                                 rjs, user)
        _project_service(request).dump_domain_in_local_mode()
        if not updated_response:
            return error(404, "ResponseNotFound", "Response could not be found.")
        return response.json(updated_response)

    @nlg_endpoints.route("/templates/<response_id>", methods=['DELETE'])
    @platform_scoped('responseTemplates.delete', allow_api_token=True)
    async def delete_template(request, response_id):
        deleted = _nlg_service(request).delete_template(response_id)
        if deleted:
            _project_service(request).dump_domain_in_local_mode()
            return response.text("", 204)
        return error(404, "ResponseNotFound", "Response could not be found.")

    return nlg_endpoints
