import json
from typing import Any, Text, Dict, Union

from sqlalchemy import Column, String, ForeignKey, Boolean, Integer, Table, Float
from sqlalchemy.orm import relationship

from rasax.community.database.base import Base

# Table for many-to-many relationship between users and roles.
user_role_mapping = Table('user_role_mapping', Base.metadata,
                          Column('username', String,
                                 ForeignKey('rasa_x_user.username')),
                          Column('role', String, ForeignKey('user_role.role')))


class User(Base):
    """Stores the Rasa X users and their credentials."""

    __tablename__ = "rasa_x_user"

    username = Column(String, primary_key=True)
    name_id = Column(String, nullable=True)
    password_hash = Column(String, nullable=True)
    team = Column(String, nullable=False)
    api_token = Column(String, nullable=False)
    project = Column(String, ForeignKey("project.project_id"))

    # Store whether the username has been deliberately assigned.
    # This property if False for SAML users who have not yet assigned
    # their own username.
    # as_dict() will return {"username": None} if username_is_assigned == False
    username_is_assigned = Column(Boolean, default=True)

    # one-to-one relationship between single-use token and user
    single_use_token = relationship("SingleUseToken",
                                    uselist=False,
                                    back_populates="user")

    roles = relationship("Role", secondary=user_role_mapping, backref="users")

    authentication_mechanism = Column(String)
    attributes = Column(String)

    def __repr__(self):
        roles = [role.role for role in self.roles]

        return "User {} (roles={}, team={})".format(self.username, roles,
                                                    self.team)

    def as_dict(self, return_api_token=False) -> Dict[Text, Any]:
        from rasax.community.services.role_service import RoleService
        role_service = RoleService()

        roles = [role.role for role in self.roles]
        user_dict = {
            "username": self.username if self.username_is_assigned else None,
            "roles": roles,
            "projects": [{"name": self.project, "roles": roles}],
            "team": self.team,
            "authentication_mechanism": self.authentication_mechanism,
            "permissions": []
        }

        if self.name_id:
            user_dict["name_id"] = self.name_id

        if self.single_use_token:
            user_dict["single_use_token"] = self.single_use_token.token
            user_dict["single_use_token_expires"] = self.single_use_token.expires

        if return_api_token:
            user_dict["api_token"] = self.api_token

        if self.attributes:
            user_dict["attributes"] = json.loads(self.attributes)

        user_dict["permissions"] = []
        for role in self.roles:
            role_permissions = [p.permission for p in role.permissions]
            user_dict["permissions"] += role_service.expand_wildcard_permissions(
                role_permissions)
        user_dict["permissions"] = list(set(user_dict["permissions"]))

        return user_dict


class Project(Base):
    """Stores different bot projects."""

    __tablename__ = "project"

    project_id = Column(String, primary_key=True)
    team = Column(String, nullable=False)
    active_model = Column(String)
    config = Column(String)
    handoff_url = Column(String)

    models = relationship("Model", cascade="all, delete-orphan",
                          back_populates="project")

    def as_dict(self) -> Dict[Text, Text]:
        return {"project_id": self.project_id, "name": self.project_id,
                "config": json.loads(self.config) if self.config else {},
                "team": self.team}

    def get_model_config(self) -> Dict[Text, Text]:
        """Return model config.

        Removes `path` key from model config, as it is unnecessary for
        for the yaml dump.
        """

        _config = json.loads(self.config)
        if "path" in _config:
            del _config["path"]

        return _config


class Environment(Base):
    """Stores the different bot environments which are available.

    E.g. a production environment and a development environment to test new
    version of the bot.
    """

    __tablename__ = "environment"

    name = Column(String, primary_key=True)
    project = Column(String, ForeignKey("project.project_id"), primary_key=True)
    url = Column(String, nullable=False)
    token = Column(String)
    db = Column(String)

    def as_dict(self) -> Dict[Text, Any]:
        return {
            self.name: {
                "url": self.url,
                "token": self.token,
                "db": self.db
            }
        }


class PlatformFeature(Base):
    """Stores whether certain feature flags are activated."""

    __tablename__ = "platform_feature"

    feature_name = Column(String, primary_key=True)
    enabled = Column(Boolean, default=False)

    def as_dict(self) -> Dict[Text, Union[Text, bool]]:
        return {
            "name": self.feature_name,
            "enabled": self.enabled
        }


class Role(Base):
    """Stores the different roles which Platform users can have."""

    __tablename__ = "user_role"

    role = Column(String, primary_key=True)
    permissions = relationship("Permission", cascade="all, delete-orphan",
                               back_populates="role")

    def as_dict(self):
        return {"role": self.role}


class Permission(Base):
    """Stores the permissions for each role."""

    __tablename__ = "permission"

    project = Column(String, ForeignKey("project.project_id"),
                     primary_key=True)
    role_id = Column(String, ForeignKey("user_role.role"), primary_key=True)
    permission = Column(String, primary_key=True)

    role = relationship("Role", back_populates="permissions")


class ChatToken(Base):
    """Stores chat_token used to share bots."""

    __tablename__ = "chat_tokens"

    token = Column(String, primary_key=True)
    bot_name = Column(String)
    description = Column(String)
    expires = Column(Integer)


class SingleUseToken(Base):
    """Stores single-use token used for enterprise single sign-on."""

    __tablename__ = "single_use_token"

    token = Column(String, primary_key=True)
    expires = Column(Float)
    username = Column(String, ForeignKey('rasa_x_user.username'))
    user = relationship("User", back_populates="single_use_token")


class LocalPassword(Base):
    """Stores the Rasa X password for use in local mode."""

    __tablename__ = "password"

    password = Column(String, primary_key=True)
